
#include "ulaw.h"

/*
 * This routine converts from linear to ulaw
 * 29 September 1989
 *
 * Craig Reese: IDA/Supercomputing Research Center
 * Joe Campbell: Department of Defense
 *
 * References:
 * 1) CCITT Recommendation G.711  (very difficult to follow)
 * 2) "A New Digital Technique for Implementation of Any 
 *     Continuous PCM Companding Law," Villeret, Michel,
 *     et al. 1973 IEEE Int. Conf. on Communications, Vol 1,
 *     1973, pg. 11.12-11.17
 * 3) MIL-STD-188-113,"Interoperability and Performance Standards
 *     for Analog-to_Digital Conversion Techniques,"
 *     17 February 1987
 *
 * Input: Signed 16 bit linear sample (Be - BIG_ENDIAN)
 * Output: 8 bit ulaw sample
 */

#define ZEROTRAP    /* turn on the trap as per the MIL-STD */
#define BIAS 0x84   /* define the add-in bias for 16 bit samples */
#define CLIP 32635

unsigned char linear2ulaw(int sample)
{
    static int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
                               4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                               5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                               5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
                               7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};
    int sign, exponent, mantissa;
    unsigned char ulawbyte;

    /** get the sample into sign-magnitude **/
    sign = (sample >> 8) & 0x80;        /* set aside the sign */
    if (sign != 0) sample = -sample;    /* get magnitude */
    if (sample > CLIP) sample = CLIP;   /* clip the magnitude */
    /** convert from 16 bit linear to ulaw **/
    sample = sample + BIAS;
    exponent = exp_lut[(sample>>7) & 0xFF];
    mantissa = (sample >> (exponent+3)) & 0x0F;
    ulawbyte = ~(sign | (exponent << 4) | mantissa);
#ifdef ZEROTRAP
    if (ulawbyte == 0 ) ulawbyte = 0x02;  /* optional CCITT trap */
#endif
    /** return the result **/
    return(ulawbyte);
}

/*
 * This routine converts from ulaw to 16 bit linear
 * 29 September 1989
 *
 * Craig Reese: IDA/Supercomputing Research Center
 *
 * References:
 * 1) CCITT Recommendation G.711  (very difficult to follow)
 * 2) MIL-STD-188-113,"Interoperability and Performance Standards
 *     for Analog-to_Digital Conversion Techniques,"
 *     17 February 1987
 *
 * Input: 8 bit ulaw sample
 * Output: signed 16 bit linear sample (Be - BIG_ENDIAN)
 */

int ulaw2linear(unsigned char ulawbyte)
{
  static int exp_lut[8]={0,132,396,924,1980,4092,8316,16764};
  int sign, exponent, mantissa, sample;

    ulawbyte = ~ulawbyte;
    sign = (ulawbyte & 0x80);
    exponent = (ulawbyte >> 4) & 0x07;
    mantissa = ulawbyte & 0x0F;
    sample = exp_lut[exponent] + (mantissa << (exponent+3));
    if (sign != 0) sample = -sample;
    return(sample);
}


/*
** This table was generated using the above function
** There shouldn't be any case where you want to use the
** function instead of this simple lookup table, but the
** function is included so that you can see the basis from
** which the numbers were derived.
*/

short _u2l[256] =
{
-32124, -31100, -30076, -29052, -28028, -27004, -25980, -24956, 
-23932, -22908, -21884, -20860, -19836, -18812, -17788, -16764, 
-15996, -15484, -14972, -14460, -13948, -13436, -12924, -12412, 
-11900, -11388, -10876, -10364, -9852, -9340, -8828, -8316, 
-7932, -7676, -7420, -7164, -6908, -6652, -6396, -6140, 
-5884, -5628, -5372, -5116, -4860, -4604, -4348, -4092, 
-3900, -3772, -3644, -3516, -3388, -3260, -3132, -3004, 
-2876, -2748, -2620, -2492, -2364, -2236, -2108, -1980, 
-1884, -1820, -1756, -1692, -1628, -1564, -1500, -1436, 
-1372, -1308, -1244, -1180, -1116, -1052, -988, -924, 
-876, -844, -812, -780, -748, -716, -684, -652, 
-620, -588, -556, -524, -492, -460, -428, -396, 
-372, -356, -340, -324, -308, -292, -276, -260, 
-244, -228, -212, -196, -180, -164, -148, -132, 
-120, -112, -104, -96, -88, -80, -72, -64, 
-56, -48, -40, -32, -24, -16, -8, 0, 
32124, 31100, 30076, 29052, 28028, 27004, 25980, 24956, 
23932, 22908, 21884, 20860, 19836, 18812, 17788, 16764, 
15996, 15484, 14972, 14460, 13948, 13436, 12924, 12412, 
11900, 11388, 10876, 10364, 9852, 9340, 8828, 8316, 
7932, 7676, 7420, 7164, 6908, 6652, 6396, 6140, 
5884, 5628, 5372, 5116, 4860, 4604, 4348, 4092, 
3900, 3772, 3644, 3516, 3388, 3260, 3132, 3004, 
2876, 2748, 2620, 2492, 2364, 2236, 2108, 1980, 
1884, 1820, 1756, 1692, 1628, 1564, 1500, 1436, 
1372, 1308, 1244, 1180, 1116, 1052, 988, 924, 
876, 844, 812, 780, 748, 716, 684, 652, 
620, 588, 556, 524, 492, 460, 428, 396, 
372, 356, 340, 324, 308, 292, 276, 260, 
244, 228, 212, 196, 180, 164, 148, 132, 
120, 112, 104, 96, 88, 80, 72, 64, 
56, 48, 40, 32, 24, 16, 8, 0
};


/*
** This version of the table was found in the 
** general sound conversion library.  It really
** isn't that good.  The source of the function used
** to generate it is unknown.

short _u2l[] =
{
 -32256, -31228, -30200, -29172, -28143, -27115, -26087, -25059,
 -24031, -23002, -21974, -20946, -19918, -18889, -17861, -16833,
 -16062, -15548, -15033, -14519, -14005, -13491, -12977, -12463,
 -11949, -11435, -10920, -10406, -9892, -9378, -8864, -8350,
 -7964, -7707, -7450, -7193, -6936, -6679, -6422, -6165,
 -5908, -5651, -5394, -5137, -4880, -4623, -4365, -4108,
 -3916, -3787, -3659, -3530, -3402, -3273, -3144, -3016,
 -2887, -2759, -2630, -2502, -2373, -2245, -2116, -1988,
 -1891, -1827, -1763, -1698, -1634, -1570, -1506, -1441,
 -1377, -1313, -1249, -1184, -1120, -1056, -992, -927,
 -879, -847, -815, -783, -751, -718, -686, -654,
 -622, -590, -558, -526, -494, -461, -429, -397,
 -373, -357, -341, -325, -309, -293, -277, -261,
 -245, -228, -212, -196, -180, -164, -148, -132,
 -120, -112, -104, -96, -88, -80, -72, -64,
 -56, -48, -40, -32, -24, -16, -8, 0,
 32256, 31228, 30200, 29172, 28143, 27115, 26087, 25059,
 24031, 23002, 21974, 20946, 19918, 18889, 17861, 16833,
 16062, 15548, 15033, 14519, 14005, 13491, 12977, 12463,
 11949, 11435, 10920, 10406, 9892, 9378, 8864, 8350,
 7964, 7707, 7450, 7193, 6936, 6679, 6422, 6165,
 5908, 5651, 5394, 5137, 4880, 4623, 4365, 4108,
 3916, 3787, 3659, 3530, 3402, 3273, 3144, 3016,
 2887, 2759, 2630, 2502, 2373, 2245, 2116, 1988,
 1891, 1827, 1763, 1698, 1634, 1570, 1506, 1441,
 1377, 1313, 1249, 1184, 1120, 1056, 992, 927,
 879, 847, 815, 783, 751, 718, 686, 654,
 622, 590, 558, 526, 494, 461, 429, 397,
 373, 357, 341, 325, 309, 293, 277, 261,
 245, 228, 212, 196, 180, 164, 148, 132,
 120, 112, 104, 96, 88, 80, 72, 64,
 56, 48, 40, 32, 24, 16, 8, 0
};
*/
